@@ -13,7 +13,7 @@
 #include	"md5/md5.h"
 #include	"auth.h"
 
-static const char rcsid[]="$Id: checkpasswordmd5.c,v 1.2 1999/12/06 13:12:28 mrsam Exp $";
+static const char rcsid[]="$Id: checkpasswordmd5.c,v 1.4 2000/06/29 01:45:10 mrsam Exp $";
 
 int authcheckpasswordmd5(const char *password, const char *encrypted_password)
 {
@@ -22,5 +22,15 @@ int authcheckpasswordmd5(const char *password, const char *encrypted_password)
 		return (strcmp(encrypted_password,
 			md5_crypt(password, encrypted_password)));
 	}
+
+#if 0
+
+	/* NOT UNTIL THIS IS TESTED */
+
+	if (strncmp(encrypted_password, "{MD5}", 5) == 0)
+	{
+		return (strcmp(encrypted_password, md5_hash(password)));
+	}
+#endif
 	return (-1);
 }
