@@ -26,7 +26,7 @@
 #include	<sys/types.h>
 #include	<sys/stat.h>
 
-static const char rcsid[]="$Id: fetch.c,v 1.6 2000/03/14 01:14:41 mrsam Exp $";
+static const char rcsid[]="$Id: fetch.c,v 1.7 2000/04/30 15:19:22 mrsam Exp $";
 
 extern int current_mailbox_ro;
 extern struct imapscaninfo current_mailbox_info;
@@ -835,7 +835,7 @@ off_t start_pos, end_pos, start_body;
 off_t nlines, nbodylines;
 size_t i,j,k,l;
 int	c;
-char	buf[BUFSIZ];
+char	buf[BUFSIZ+2];
 int	goodheader;
 unsigned long skipping;
 unsigned long cnt;
@@ -882,6 +882,8 @@ char	*p;
 				buf[k]=':';
 			}
 		}
+		else if (buf[j] == '\n')
+			goodheader=0;
 
 		for (k=j; k<i; k++)
 			if (buf[k] == '\n')
@@ -898,6 +900,8 @@ char	*p;
 		j=k;
 	}
 
+	buf[l++]='\n';	/* Always append a blank line */
+
 	cnt=l;
 	for (i=0; i<l; i++)
 		if (buf[i] == '\n')	++cnt;
@@ -913,7 +917,7 @@ char	*p;
 	}
 
 	writes("{");
-	writen(cnt+2);	/* BUG */
+	writen(cnt);
 	writes("}\r\n");
 	p=buf;
 	while (skipping)
@@ -953,7 +957,6 @@ char	*p;
 		p += i;
 		cnt -= i;
 	}
-	writes("\r\n");	/* BUG */
 }
 
 struct fetchheaderinfo {
@@ -1037,6 +1040,9 @@ struct	fetchheaderinfo finfo;
 			if (buf1[0] != '\n' && buf1[0] != '\r' &&
 				!isspace((int)(unsigned char)buf1[0]))
 				goodheader= (*headerfunc)(fi, buf1);
+			else if (buf1[0] == '\n')
+				goodheader=0;
+
 			if (!goodheader)
 			{
 				while (left)
