@@ -58,6 +58,7 @@
 #include	"searchinfo.h"
 #include	"storeinfo.h"
 #include	"mailboxlist.h"
+#include	"thread.h"
 
 #include	"authlib/authmod.h"
 
@@ -69,7 +70,7 @@
 
 #define	SHARED	"shared"
 
-static const char rcsid[]="$Id: imapd.c,v 1.28 2000/02/28 05:16:52 mrsam Exp $";
+static const char rcsid[]="$Id: imapd.c,v 1.35 2000/06/17 14:44:49 mrsam Exp $";
 
 extern time_t rfc822_parsedt(const char *);
 extern void fetchflags(unsigned long);
@@ -246,9 +247,6 @@ char	*arg;
 			arg=maildir_folderdir(0, arg);
 	}
 
-	if (arg)
-		arg=my_strdup(arg);
-
 	if (!arg && errno != EINVAL)
 	{
 		free(p);
@@ -553,6 +551,10 @@ int	lastnl;
 
 	if (fflush(fp) || ferror(fp))
 	{
+		fprintf(stderr,
+                        "ERR: error storing a message, user=%s, errno=%d\n",
+                                getenv("AUTHENTICATED"), errno);
+
 		fclose(fp);
 		close(fd);
 		unlink(tmpname);
@@ -939,6 +941,11 @@ static void expunge()
 unsigned long j;
 struct imapflags flags;
 char	*old_name;
+int	move_to_trash=0;
+
+	old_name=getenv("IMAP_MOVE_EXPUNGE_TO_TRASH");
+	if (old_name && atoi(old_name) && !is_trash(current_mailbox))
+		move_to_trash=1;
 
 	for (j=0; j < current_mailbox_info.nmessages; j++)
 	{
@@ -955,6 +962,32 @@ char	*old_name;
 		{
 			maildir_unlinksharedmsg(old_name);
 		}
+		else if (move_to_trash)
+		{
+		char	*new_name;
+		char	*p;
+
+			new_name=alloc_filename("." TRASH,
+				current_mailbox_info.msgs[j].filename);
+
+			/* Don't mark it as deleted in the Trash */
+
+			if ((p=strrchr(new_name, '/')) &&
+				(p=strrchr(p, ':')) &&
+				strncmp(p, ":2,", 3) == 0 &&
+				(p=strchr(p, 'T')))
+				while ((*p=p[1]) != 0)
+					++p;
+			if (rename(old_name, new_name))
+			{
+				writes("* NO [ALERT] Error while moving ");
+				writes(old_name);
+				writes(" to ");
+				writes(new_name);
+				writes("\r\n");
+			}
+			free(new_name);
+		}
 		else
 		{
 			unlink(old_name);
@@ -1677,6 +1710,8 @@ int	uid=0;
 		if (strcmp(curtoken->tokenbuf, "COPY") &&
 			strcmp(curtoken->tokenbuf, "FETCH") &&
 			strcmp(curtoken->tokenbuf, "SEARCH") &&
+			strcmp(curtoken->tokenbuf, "THREAD") &&
+			strcmp(curtoken->tokenbuf, "SORT") &&
 			strcmp(curtoken->tokenbuf, "STORE"))
 			return (-1);
 	}
@@ -1806,8 +1841,14 @@ int	uid=0;
 		if (currenttoken()->tokentype != IT_EOL)
 		{
 			if (charset)	free(charset);
+			free_search(sihead);
 			return (-1);
 		}
+		if (charset && strcasecmp(charset, "US-ASCII") == 0)
+		{
+			free(charset);
+			charset=0;
+		}
 		if (charset)
 		{
 			writes(tag);
@@ -1816,8 +1857,6 @@ int	uid=0;
 			free_search(sihead);
 			return (0);
 		}
-		free(charset);
-
 #if 0
 		writes("* OK ");
 		debug_search(si);
@@ -1836,6 +1875,246 @@ int	uid=0;
 		return (0);
 	}
 
+	if (strcmp(curtoken->tokenbuf, "THREAD") == 0)
+	{
+	char *charset=0;
+	struct searchinfo *si, *sihead;
+	unsigned long i;
+
+		/* The following jazz is mainly for future extensions */
+
+	void (*thread_func)(struct searchinfo *, struct searchinfo *, int);
+	search_type thread_type;
+
+		{
+		const char *p=getenv("IMAP_DISABLETHREADSORT");
+		int n= p ? atoi(p):0;
+
+			if (n > 0)
+			{
+				writes(tag);
+				writes(" NO This command is disabled by the system administrator.\r\n");
+				return (0);
+			}
+		}
+
+		curtoken=nexttoken();
+		if (curtoken->tokentype != IT_ATOM &&
+			curtoken->tokentype != IT_QUOTED_STRING)
+			return (-1);
+
+		if (strcmp(curtoken->tokenbuf, "ORDEREDSUBJECT") == 0)
+		{
+			thread_func=dothreadorderedsubj;
+			thread_type=search_orderedsubj;
+		}
+		else
+		{
+			return (-1);
+		}
+
+		curtoken=nexttoken();
+		if (curtoken->tokentype != IT_ATOM &&
+			curtoken->tokentype != IT_QUOTED_STRING)
+			return (-1);
+
+
+		charset=my_strdup(curtoken->tokenbuf);
+		curtoken=nexttoken();
+
+		if ((si=alloc_parsesearch(&sihead)) == 0)
+		{
+			if (charset)	free(charset);
+			return (-1);
+		}
+
+		si=alloc_searchextra(si, &sihead, thread_type);
+
+		if (currenttoken()->tokentype != IT_EOL)
+		{
+			if (charset)	free(charset);
+			free_search(sihead);
+			return (-1);
+		}
+		if (charset && strcasecmp(charset, "US-ASCII") == 0)
+		{
+			free(charset);
+			charset=0;
+		}
+		if (charset)
+		{
+			writes(tag);
+			writes(" NO Can't thread with that character set.\r\n");
+			if (charset)	free(charset);
+			free_search(sihead);
+			return (0);
+		}
+
+		writes("* THREAD ");
+		(*thread_func)(si, sihead, uid);
+		writes("\r\n");
+
+		for (i=0; i<current_mailbox_info.nmessages; i++)
+			if (current_mailbox_info.msgs[i].changedflags)
+				fetchflags(i);
+		writes(tag);
+		writes(" OK THREAD done.\r\n");
+		free_search(sihead);
+		return (0);
+	}
+
+	if (strcmp(curtoken->tokenbuf, "SORT") == 0)
+	{
+	char *charset=0;
+	struct searchinfo *si, *sihead;
+	unsigned long i;
+	struct temp_sort_stack *ts=0;
+
+		{
+		const char *p=getenv("IMAP_DISABLETHREADSORT");
+		int n= p ? atoi(p):0;
+
+			if (n > 0)
+			{
+				writes(tag);
+				writes(" NO This command is disabled by the system administrator.\r\n");
+				return (0);
+			}
+		}
+
+		curtoken=nexttoken();
+		if (curtoken->tokentype != IT_LPAREN)	return (-1);
+		while ((curtoken=nexttoken())->tokentype != IT_RPAREN)
+		{
+		search_type st;
+		struct temp_sort_stack *newts;
+
+			if (curtoken->tokentype != IT_ATOM &&
+				curtoken->tokentype != IT_QUOTED_STRING)
+			{
+				free_temp_sort_stack(ts);
+				return (-1);
+			}
+
+			if (strcmp(curtoken->tokenbuf, "SUBJECT") == 0)
+			{
+				st=search_orderedsubj;
+			}
+			else if (strcmp(curtoken->tokenbuf, "ARRIVAL") == 0)
+			{
+				st=search_arrival;
+			}
+			else if (strcmp(curtoken->tokenbuf, "CC") == 0)
+			{
+				st=search_cc;
+			}
+			else if (strcmp(curtoken->tokenbuf, "DATE") == 0)
+			{
+				st=search_date;
+			}
+			else if (strcmp(curtoken->tokenbuf, "FROM") == 0)
+			{
+				st=search_from;
+			}
+			else if (strcmp(curtoken->tokenbuf, "REVERSE") == 0)
+			{
+				st=search_reverse;
+			}
+			else if (strcmp(curtoken->tokenbuf, "SIZE") == 0)
+			{
+				st=search_size;
+			}
+			else if (strcmp(curtoken->tokenbuf, "TO") == 0)
+			{
+				st=search_to;
+			}
+			else
+			{
+				free_temp_sort_stack(ts);
+				return (-1);
+			}
+
+			newts=(struct temp_sort_stack *)malloc(
+				sizeof(struct temp_sort_stack *));
+			if (!newts)	write_error_exit(0);
+			newts->next=ts;
+			newts->type=st;
+			ts=newts;
+		}
+
+		if (ts == 0	/* No criteria */
+			|| ts->type == search_reverse)
+				/* Can't end with the REVERSE keyword */
+		{
+			free_temp_sort_stack(ts);
+			return (-1);
+		}
+
+		curtoken=nexttoken();
+		if (curtoken->tokentype != IT_ATOM &&
+			curtoken->tokentype != IT_QUOTED_STRING)
+		{
+			free_temp_sort_stack(ts);
+			return (-1);
+		}
+
+		charset=my_strdup(curtoken->tokenbuf);
+		curtoken=nexttoken();
+
+		if ((si=alloc_parsesearch(&sihead)) == 0)
+		{
+			if (charset)	free(charset);
+			free_temp_sort_stack(ts);
+			return (-1);
+		}
+
+		while (ts)
+		{
+		struct temp_sort_stack *cts=ts;
+
+			ts=cts->next;
+			si=alloc_searchextra(si, &sihead, cts->type);
+			free(cts);
+		}
+
+		if (currenttoken()->tokentype != IT_EOL)
+		{
+			if (charset)	free(charset);
+			free_search(sihead);
+			return (-1);
+		}
+		if (charset && strcasecmp(charset, "US-ASCII") == 0)
+		{
+			free(charset);
+			charset=0;
+		}
+		if (charset && strcasecmp(charset, "UTF-8") == 0)
+		{
+			free(charset);
+			charset=0;
+		}
+		if (charset)
+		{
+			writes(tag);
+			writes(" NO Can't sort with that character set.\r\n");
+			if (charset)	free(charset);
+			free_search(sihead);
+			return (0);
+		}
+
+		writes("* SORT");
+		dosortmsgs(si, sihead, uid);
+		writes("\r\n");
+
+		for (i=0; i<current_mailbox_info.nmessages; i++)
+			if (current_mailbox_info.msgs[i].changedflags)
+				fetchflags(i);
+		writes(tag);
+		writes(" OK SORT done.\r\n");
+		free_search(sihead);
+		return (0);
+	}
+
 	if (strcmp(curtoken->tokenbuf, "CHECK") == 0)
 	{
 		if (nexttoken()->tokentype != IT_EOL)	return (-1);
@@ -1879,16 +2158,28 @@ int	uid=0;
 		}
 
 		mailbox=decode_valid_mailbox(curtoken->tokenbuf, 1);
-		if (mailbox && *mailbox == 0)
-		{
-			free(mailbox);
-			mailbox=0;
-		}
+
 		if (!mailbox)
 		{
 			free(msgset);
+
+			mailbox=valid_mailbox_name(curtoken->tokenbuf, 1);
+			if (mailbox && *mailbox == 0)
+			{
+				free(mailbox);
+				mailbox=0;
+			}
+			if (mailbox && nexttoken()->tokentype == IT_EOL)
+			{
+				free(mailbox);
+				writes(tag);
+				writes(" NO [TRYCREATE] Mailbox does not exist.\r\n");
+				return (0);
+			}
+			if (mailbox)	free(mailbox);
 			return (-1);
 		}
+
 		if (nexttoken()->tokentype != IT_EOL)
 		{
 			free(msgset);
