@@ -19,12 +19,13 @@
 #include	<sys/stat.h>
 #include	"rfc822/rfc822.h"
 #include	"rfc2045/rfc2045.h"
+#include	"numlib/numlib.h"
 #include	"searchinfo.h"
 #include	"imapwrite.h"
 #include	"imaptoken.h"
 #include	"imapscanclient.h"
 
-static const char rcsid[]="$Id: search.c,v 1.3 1999/12/06 13:33:00 mrsam Exp $";
+static const char rcsid[]="$Id: search.c,v 1.5 2000/06/09 12:29:14 mrsam Exp $";
 
 extern time_t rfc822_parsedt(const char *);
 extern struct imapscaninfo current_mailbox_info;
@@ -53,8 +54,33 @@ static void searcherror(const char *errmsg,
                 errmsg, (unsigned)getuid(), mi->filename);
 }
 
+static void search_callback(struct searchinfo *, struct searchinfo *, int,
+	unsigned long, void *);
+
+/*
+**	search_internal() does the main heavylifting of searching the
+**	maildir for qualifying messages.  It calls a callback function
+**	when a matching message is found.
+**
+**	For a plain SEARCH, the callback function merely prints the message
+**	number.
+*/
 
 void dosearch(struct searchinfo *si, struct searchinfo *sihead, int isuid)
+{
+	search_internal(si, sihead, isuid, search_callback, 0);
+}
+
+static void search_callback(struct searchinfo *si, struct searchinfo *sihead,
+	int isuid, unsigned long i, void *dummy)
+{
+	writes(" ");
+	writen(isuid ? current_mailbox_info.msgs[i].uid:i+1);
+}
+
+void search_internal(struct searchinfo *si, struct searchinfo *sihead,
+	int isuid, void (*callback_func)(struct searchinfo *,
+		struct searchinfo *, int, unsigned long, void *), void *voidarg)
 {
 unsigned long i;
 struct	searchinfo *p;
@@ -111,8 +137,7 @@ int	rc;
 
 		if (rc > 0)
 		{
-			writes(" ");
-			writen(isuid ? current_mailbox_info.msgs[i].uid:i+1);
+			(*callback_func)(si, sihead, isuid, i, voidarg);
 		}
 		fclose(fp);
 		close(fd);
@@ -201,6 +226,17 @@ char	buf1[60], buf2[80];
 	return (rfc822_parsedt(buf2));
 }
 
+static char *timestamp_for_sorting(time_t t)
+{
+struct tm *tm=localtime(&t);
+char	buf[200];
+
+	buf[0]=0;
+	if ( strftime(buf, sizeof(buf), "%Y.%m.%d.%H.%M.%S", tm) == 0)
+		buf[0]=0;
+	return (my_strdup(buf));
+}
+
 /* Evaluate non-content search nodes */
 
 static void fill_search_quick(struct searchinfo *p,
@@ -284,10 +320,52 @@ static void fill_search_quick(struct searchinfo *p,
 		}
 		break;
 	case search_uid:
-		break;
 		if (is_in_set(p->as, current_mailbox_info.msgs[msgnum].uid))
 			p->value=1;
 		break;
+	case search_orderedsubj:
+	case search_arrival:
+	case search_cc:
+	case search_date:
+	case search_from:
+	case search_reverse:
+	case search_size:
+	case search_to:
+
+		/* DUMMY nodes for SORT/THREAD.  Make sure that the
+		** dummy node is CLEARed */
+
+		if (p->as)
+		{
+			free(p->as);
+			p->as=0;
+		}
+
+		switch (p->type)	{
+		case search_arrival:
+			p->as=timestamp_for_sorting(stat_buf->st_mtime);
+			p->value=1;
+			break;
+		case search_size:
+			{
+			char	buf[NUMBUFSIZE], buf2[NUMBUFSIZE];
+			char *q;
+
+				str_size_t(stat_buf->st_size, buf);
+				sprintf(buf2, "%*s", sizeof(buf2)-1, buf);
+				for (q=buf2; *q == ' '; *q++='0')
+					;
+				p->as=my_strdup(buf2);
+				p->value=1;
+			}
+			break;
+		case search_reverse:
+			p->value=1;
+			break;
+		default:
+			break;
+		}
+		break;
 	default:
 		break;
 	}
@@ -304,6 +382,16 @@ static int search_evaluate(struct searchinfo *si)
 int	rc, rc2;
 
 	switch (si->type)	{
+	case search_orderedsubj:	/* DUMMY for THREAD and SORT */
+        case search_arrival:
+        case search_cc:
+        case search_date:
+        case search_from:
+        case search_reverse:
+        case search_size:
+        case search_to:
+		if (si->value < 0)	return (-1);
+		return (search_evaluate(si->a));
 	case search_not:
 		rc=search_evaluate(si->a);
 		if (rc >= 0)	rc= 1-rc;
@@ -329,6 +417,89 @@ int	rc, rc2;
 	return (rc);
 }
 
+/* Remove artifacts from the subject header */
+
+static void stripsubj(char *s)
+{
+char	*p;
+char	*q;
+
+	for (p=s; *p; p++)
+		;
+	while (p > s)
+	{
+		if ( isspace((int)(unsigned char)p[-1]))
+		{
+			--p;
+			continue;
+		}
+		if (p-s >= 3 && strcmp(p-3, "(FWD)") == 0)
+		{
+			p -= 3;
+			continue;
+		}
+		break;
+	}
+	*p=0;
+
+	for (p=s; *p; )
+	{
+		for (;;)
+		{
+			if (isspace((int)(unsigned char)*p))
+			{
+				++p;
+				continue;
+			}
+			if (strncmp(p, "RE", 2) == 0)
+			{
+				q=p+2;
+			}
+			else if (strncmp(p, "FWD", 3) == 0)
+			{
+				q=p+3;
+			}
+			else if (strncmp(p, "FW", 2) == 0)
+			{
+				q=p+2;
+			}
+			else	break;
+
+			if (*q == '[')
+			{
+				++q;
+				if (!isdigit((int)(unsigned char)*q))
+					break;
+				do
+				{
+					++q;
+				} while (isdigit((int)(unsigned char)*q));
+				if (*q != ']')	break;
+				++q;
+			}
+			if (*q++ != ':')	break;
+
+			while (*q && isspace((int)(unsigned char)*q))
+				++q;
+			p=q;
+		}
+
+		if (*p != '[')	break;
+		for (q=p; *q; q++)
+			if (*q == ']')
+				break;
+		if (*q != ']')	break;
+		++q;
+		while (*q && isspace((int)(unsigned char)*q))
+			++q;
+		p=q;
+	}
+
+	q=s;
+	while ( (*q++ = *p++) != 0)
+		;
+}
+
 /* ------- header search -------- */
 
 static void fill_search_header_recursive(struct searchinfo *,
@@ -433,6 +604,72 @@ struct searchinfo *sip;
 				break;
 			}
 
+			if (sip->type == search_cc && strcmp(p, "CC") == 0 &&
+				sip->as == 0)
+			{
+			struct rfc822t *t=rfc822t_alloc(q, 0);
+			struct rfc822a *a=t ? rfc822a_alloc(t):0;
+
+				sip->as= a && a->naddrs > 0 ?
+					rfc822_getaddr(a, 0):0;
+				if (a)	rfc822a_free(a);
+				if (t)	rfc822t_free(t);
+			}
+
+			if (sip->type == search_date && strcmp(p, "DATE") == 0
+				&& sip->as == 0)
+			{
+			time_t msg_time=rfc822_parsedt(q);
+
+				sip->as=timestamp_for_sorting(msg_time);
+			}
+
+			if (sip->type == search_from && strcmp(p, "FROM") == 0
+				&& sip->as == 0)
+			{
+			struct rfc822t *t=rfc822t_alloc(q, 0);
+			struct rfc822a *a=t ? rfc822a_alloc(t):0;
+
+				sip->as= a && a->naddrs > 0 ?
+					rfc822_getaddr(a, 0):0;
+				if (a)	rfc822a_free(a);
+				if (t)	rfc822t_free(t);
+			}
+
+			if (sip->type == search_to && strcmp(p, "TO") == 0 &&
+				sip->as == 0)
+			{
+			struct rfc822t *t=rfc822t_alloc(q, 0);
+			struct rfc822a *a=t ? rfc822a_alloc(t):0;
+
+				sip->as= a && a->naddrs > 0 ?
+					rfc822_getaddr(a, 0):0;
+				if (a)	rfc822a_free(a);
+				if (t)	rfc822t_free(t);
+			}
+
+			switch (sip->type)	{
+			case search_cc:
+			case search_date:
+			case search_from:
+			case search_to:
+				sip->value=1;
+				break;
+			default:
+				break;
+			}
+
+			if (sip->type == search_orderedsubj &&
+				strcmp(p, "SUBJECT") == 0 && sip->as == 0)
+			{
+			/* Dummy node used to suck in the Subject:
+			** header for the THREAD or SORT command.
+			*/
+				sip->as=my_strdup(q);
+				sip->value=1;
+				stripsubj(sip->as);
+			}
+
 			if (sip->type != search_header ||
 				sip->value > 0)	continue;
 
