@@ -1,5 +1,5 @@
 /*
-** Copyright 1998 - 1999 Double Precision, Inc.  See COPYING for
+** Copyright 1998 - 2000 Double Precision, Inc.  See COPYING for
 ** distribution information.
 */
 
@@ -35,7 +35,7 @@ int gethostname(const char *, size_t);
 #endif
 
 
-static const char rcsid[]="$Id: reformime.c,v 1.20 2000/03/22 05:21:10 mrsam Exp $";
+static const char rcsid[]="$Id: reformime.c,v 1.23 2000/05/23 17:55:02 mrsam Exp $";
 
 void rfc2045_error(const char *errmsg)
 {
@@ -227,8 +227,8 @@ const char *p;
 	printf("starting-pos: %lu\n", (unsigned long)start);
 	printf("starting-pos-body: %lu\n", (unsigned long)body);
 	printf("ending-pos: %lu\n", (unsigned long)end);
-	printf("line-count: %lu\n", nlines);
-	printf("body-line-count: %lu\n", nbodylines);
+	printf("line-count: %lu\n", (unsigned long)nlines);
+	printf("body-line-count: %lu\n", (unsigned long)nbodylines);
 }
 
 static void do_print_info_multiple(struct rfc2045 *p, struct rfc2045id *id,
@@ -349,7 +349,7 @@ char	*dyn_disp_name=0;
 	if (pfix)	strcpy(p, pfix);
 	q=p+strlen(p);
 	for (strcpy(q, disposition_filename_s); *q; q++)
-		if (!isalnum(*q) && *q != '.')
+		if (!isalnum(*q) && *q != '.' && *q != '-')
 			*q='_';
 
 	if (dyn_disp_name)	free(dyn_disp_name);
@@ -576,7 +576,7 @@ char *p, *q;
 	free(addr);
 }
 
-static void dsn(struct rfc2045 *p)
+static void dsn(struct rfc2045 *p, int do_orig)
 {
 const char *content_type_s;
 const char *content_transfer_encoding_s;
@@ -589,6 +589,7 @@ unsigned i;
 int	ch;
 char *recip;
 char *action;
+char *orecip;
 
 	rfc2045_mimeinfo(p, &content_type_s, &content_transfer_encoding_s,
 		&charset_s);
@@ -608,6 +609,7 @@ char *action;
 
 	i=0;
 	recip=0;
+	orecip=0;
 	action=0;
 	while (start_body < end_pos)
 	{
@@ -623,6 +625,12 @@ char *action;
 		buf[i-1]=0;
 		if (buf[0] == 0)
 		{
+			if (orecip)
+			{
+				if (recip)	free(recip);
+				recip=orecip;
+				orecip=0;
+			}
 			print_dsn_recip(recip, action);
 			recip=0;
 			action=0;
@@ -637,10 +645,20 @@ char *action;
 				exit(2);
 			}
 		}
+		if (strncasecmp(buf, "Original-Recipient:", 19) == 0 &&
+			orecip == 0 && do_orig)
+		{
+			orecip=strdup(buf+19);
+			if (!orecip)
+			{
+				perror("strdup");
+				exit(2);
+			}
+		}
 		if (strncasecmp(buf, "Action:", 7) == 0 && action == 0)
 		{
 			action=strdup(buf+7);
-			if (!recip)
+			if (!action)
 			{
 				perror("strdup");
 				exit(2);
@@ -648,6 +666,12 @@ char *action;
 		}
 		i=0;
 	}
+	if (orecip)
+	{
+		if (recip)	free(recip);
+		recip=orecip;
+		orecip=0;
+	}
 	print_dsn_recip(recip, action);
 }
 
@@ -820,6 +844,9 @@ const char *extract_filename=0;
 		case 'd':
 			dodsn=1;
 			break;
+		case 'D':
+			dodsn=2;
+			break;
 		case 'x':
 			do_extract=extract_file;
 			if (optarg)
@@ -848,7 +875,7 @@ const char *extract_filename=0;
 	else if (dorewrite)
 		rewrite(p, rwmode);
 	else if (dodsn)
-		dsn(p);
+		dsn(p, dodsn == 2);
 	else if (do_extract)
 		extract_section(p, mimesection, extract_filename,
 			argc-argn, argv+argn, do_extract);
