@@ -38,6 +38,9 @@
 #if	HAVE_FCNTL_H
 #include	<fcntl.h>
 #endif
+#if	HAVE_SYS_IOCTL_H
+#include	<sys/ioctl.h>
+#endif
 #if HAVE_UNISTD_H
 #include	<unistd.h>
 #endif
@@ -49,7 +52,7 @@
 
 #include	<netdb.h>
 
-static const char rcsid[]="$Id: tcpd.c,v 1.11 2000/03/29 13:15:17 mrsam Exp $";
+static const char rcsid[]="$Id: tcpd.c,v 1.16 2000/06/04 18:58:21 mrsam Exp $";
 
 static const char *accessarg=0;
 static const char *listenarg=0;
@@ -99,7 +102,7 @@ static int fd;
 
 static int nprocs, maxperc, maxperip;
 static pid_t *pids;
-static struct sockaddr *addrs;
+static RFC1035_ADDR *addrs;
 
 static int sighup_received=0;
 
@@ -146,7 +149,7 @@ struct blocklist_s **blptr;
 	if (newbl->var && (p=strchr(newbl->var, '/')) != 0)
 	{
 		*p++=0;
-		newbl->ia.s_addr=rfc1035_aton(p);
+		rfc1035_aton_ipv4(p, &newbl->ia);
 	}
 }
 
@@ -264,7 +267,20 @@ static int init(int argc, char **argv)
 {
 int	argn;
 int	port;
-struct	sockaddr_in sin;
+
+RFC1035_ADDR addr;
+RFC1035_NETADDR  netaddr;
+const struct sockaddr *sinaddr;
+int	sinaddrlen;
+
+#if	RFC1035_IPV6
+struct sockaddr_in6	sin6;
+#endif
+
+struct sockaddr_in	sin4;
+
+int	af;
+int	rc;
 struct	group *gr;
 int	i;
 gid_t	gid=0;
@@ -300,35 +316,72 @@ int	forced=0;
 		port=htons(port);
 	}
 
-	sin.sin_family=AF_INET;
-	sin.sin_port=port;
-	if (ipaddrarg == 0)
-	{
-		sin.sin_addr.s_addr=htonl(INADDR_ANY);
-	}
-	else
-	{
-		sin.sin_addr.s_addr=rfc1035_aton(ipaddrarg);
-		if ( (unsigned long)sin.sin_addr.s_addr == (unsigned long)-1)
-		{
-			fprintf(stderr,"Invalid IP address: %s\n", ipaddrarg);
-			return (-1);
-		}
-	}
+	/* Create an IPv6 or an IPv4 socket */
 
-	if ((fd=socket(PF_INET, SOCK_STREAM, 0)) < 0)
+	if ((fd=rfc1035_mksocket(SOCK_STREAM, 0, &af)) < 0)
 	{
 		perror("socket");
 		return (-1);
 	}
 
+	/* Figure out what to bind based on what socket we created */
+
+	if (ipaddrarg && strcmp(ipaddrarg, "0"))
+	{
+		if (rfc1035_aton(ipaddrarg, &addr) < 0)
+		{
+			fprintf(stderr,"Invalid IP address: %s\n", ipaddrarg);
+			close(fd);
+			return (-1);
+		}
+
+		if (rfc1035_mkaddress(af, &netaddr, &addr, port, &sinaddr,
+				&sinaddrlen))
+		{
+			fprintf(stderr,"Unable to bind IP address: %s\n",
+				ipaddrarg);
+			close(fd);
+			return (-1);
+		}
+	}
+	else	/* Bind default address */
+	{
+#if	RFC1035_IPV6
+
+		if (af == AF_INET6)
+		{
+			memset(&sin6, 0, sizeof(sin6));
+			sin6.sin6_family=AF_INET6;
+			sin6.sin6_addr=in6addr_any;
+			sin6.sin6_port=port;
+			sinaddr=(const struct sockaddr *)&sin6;
+			sinaddrlen=sizeof(sin6);
+		}
+		else
+#endif
+		if (af == AF_INET)
+		{
+			sin4.sin_family=AF_INET;
+			sin4.sin_addr.s_addr=INADDR_ANY;
+			sin4.sin_port=port;
+			sinaddr=(const struct sockaddr *)&sin4;
+			sinaddrlen=sizeof(sin4);
+		}
+		else
+		{
+			errno=EAFNOSUPPORT;
+			rc= -1;
+		}
+	}
+
 	signal(SIGINT, sigexit);
 	signal(SIGHUP, sighup);
 	signal(SIGTERM, sigexit);
 
-	while (sox_bind(fd, (struct sockaddr *)&sin, sizeof(sin)) < 0)
+	while (sox_bind(fd, (struct sockaddr *)sinaddr, sinaddrlen) < 0)
 	{
 	int	fd2;
+	int	dummy;
 
 		perror("bind");
 		if (!forcebindarg || errno != EADDRINUSE)
@@ -339,14 +392,16 @@ int	forced=0;
 
 		/* Poke around */
 
-		if ((fd2=socket(PF_INET, SOCK_STREAM, 0)) < 0)
+		if ((fd2=rfc1035_mksocket(SOCK_STREAM, 0, &dummy)) < 0)
+			/* Better get same socket as fd */
 		{
 			perror("socket");
 			sox_close(fd);
 			return (-1);
 		}
 
-		if (sox_connect(fd2, (struct sockaddr *)&sin, sizeof(sin)) == 0)
+		if (sox_connect(fd2, (struct sockaddr *)sinaddr,
+			sinaddrlen) == 0)
 		{
 			sox_close(fd2);
 			sox_close(fd);
@@ -526,7 +581,7 @@ int	forced=0;
 	return (argn);
 }
 
-static void run(int, struct sockaddr_in *, const char *, char **);
+static void run(int, const RFC1035_ADDR *, int, const char *, char **);
 
 static void doreap(pid_t p, int wait_stat)
 {
@@ -549,20 +604,63 @@ static RETSIGTYPE childsig(int signum)
 #endif
 }
 
-static int allowaccess(struct sockaddr_in *sin)
+static int doallowaccess(char *);
+
+#if	RFC1035_IPV6
+
+static int allowaccess(const RFC1035_ADDR *sin)
+{
+char	buf[RFC1035_MAXNAMESIZE+1];
+char	*q;
+int	i;
+
+	if (IN6_IS_ADDR_V4MAPPED(sin))
+	{
+	const char *p=inet_ntop(AF_INET6, sin, buf, sizeof(buf));
+
+		if (p && (q=strrchr(buf, ':')) != 0)
+			return (doallowaccess(q+1));
+		return (1);
+	}
+	q=buf;
+	for (i=0; i<sizeof(*sin); i += 2)
+	{
+#define	B(i) ((unsigned long)((unsigned char *)sin)[i])
+	unsigned long n=(B(i) << 8) | B(i+1);
+#undef	B
+		sprintf(q, ":%04lx", n);
+		q += 5;
+	}
+	*q=0;
+	return (doallowaccess(buf));
+}
+
+#else
+static int allowaccess(const RFC1035_ADDR *sin)
 {
 char	buf[RFC1035_NTOABUFSIZE];
+
+	rfc1035_ntoa(sin, buf);
+	return (doallowaccess(buf));
+}
+#endif
+
+static int doallowaccess(char *buf)
+{
 char	*accessptr;
 char *p, *q, *r;
 int	quote=0;
 int	l;
 
 	if (accessarg == 0)	return (1);
-	rfc1035_ntoa(sin->sin_addr, buf);
 
-	while ((accessptr=chkaccess(buf)) == 0)
+	while ((accessptr= *buf ? chkaccess(buf):0) == 0)
 	{
-		if ((accessptr=strrchr(buf, '.')) == 0)
+		if ((accessptr=strrchr(buf, '.')) == 0
+#if RFC1035_IPV6
+			&& (accessptr=strrchr(buf, ':')) == 0
+#endif
+			)
 		{
 			if ((accessptr=chkaccess("*")) != 0)
 				break;
@@ -683,12 +781,26 @@ int	pidptr;
 #if	HAVE_SETPGID
 	setpgid(0, 0);
 #endif
+#endif
+#ifdef  TIOCNOTTY
+
+	{
+	int fd=open("/dev/tty", O_RDWR);
+
+		if (fd >= 0)
+		{
+			ioctl(fd, TIOCNOTTY, 0);
+			close(fd);
+		}
+	}
 #endif
 
 	for (;;)
 	{
 	int	sockfd;
-	struct	sockaddr_in sin;
+	RFC1035_NETADDR	sin;
+	RFC1035_ADDR addr;
+	int	addrport;
 	int	sinl;
 #ifdef	SOL_LINGER
 	struct	linger l;
@@ -718,7 +830,8 @@ int	pidptr;
 				perror("accept");
 		}
 
-		if (sinl < sizeof(int) || sin.sin_family != AF_INET)
+		if (rfc1035_sockaddrip(&sin, sinl, &addr)
+			|| rfc1035_sockaddrport(&sin, sinl, &addrport))
 		{
 			sox_close(sockfd);
 			continue;
@@ -767,16 +880,17 @@ int	pidptr;
 			wait_restore(childsig);
 			sox_close(fd);
 
-			if (allowaccess(&sin) == 0)
+			if (allowaccess(&addr) == 0)
 			{
 				sox_close(sockfd);
 				_exit(0);
 			}
-			run(sockfd, &sin, argv[argn], ptrs);
+
+			run(sockfd, &addr, addrport, argv[argn], ptrs);
 		}
 		pids[pidptr]=p;
 
-		memcpy(addrs+pidptr, &sin, sizeof(sin));
+		memcpy(addrs+pidptr, &addr, sizeof(addr));
 		sox_close(sockfd);
 		wait_clear(childsig);
 	}
@@ -799,26 +913,28 @@ char	*p=malloc(strlen(name)+strlen(val)+2);
 ** backwards and forwards.
 */
 
-static void ip2host(struct in_addr in, const char *env)
+static void ip2host(const RFC1035_ADDR *addr, const char *env)
 {
 const char *remotehost="softdnserr";
 char	buf[RFC1035_MAXNAMESIZE+1];
 
 	if (nodnslookup)	return;
 
+	rfc1035_ntoa(addr, buf);
+
 #if TCPDUSERFC1035
 
-	if (rfc1035_ptr(&rfc1035_default_resolver, in, buf) != 0)
+	if (rfc1035_ptr(&rfc1035_default_resolver, addr, buf) != 0)
 	{
 		if (errno == ENOENT)
 			remotehost=0;
 	}
 	else
 	{
-	struct in_addr *ias;
+	RFC1035_ADDR *ias;
 	unsigned nias, n;
 
-		if (rfc1035_a(&rfc1035_default_resolver, buf, &ias, &nias) < 0)
+		if (rfc1035_a(&rfc1035_default_resolver, buf, &ias, &nias) != 0)
 		{
 			if (errno == ENOENT)
 				remotehost=0;
@@ -827,16 +943,37 @@ char	buf[RFC1035_MAXNAMESIZE+1];
 		{
 			remotehost=0;
 			for (n=0; n<nias; n++)
-				if (ias[n].s_addr == in.s_addr)
+			{
+			char	a[RFC1035_MAXNAMESIZE];
+			char	b[RFC1035_MAXNAMESIZE];
+
+				rfc1035_ntoa(&ias[n], a);
+				rfc1035_ntoa(addr, b);
+
+				if (strcmp(a, b) == 0)
+				{
 					remotehost=buf;
+				}
+			}
 		}
 	}
 #else
 
 	{
-	struct hostent *he=gethostbyaddr( (char *)&in, sizeof(in), AF_INET);
+	struct hostent *he;
 	unsigned n;
+	struct	in_addr in;
 
+#if	RFC1035_IPV6
+
+		if (IN6_IS_ADDR_V4MAPPED(sin))
+			memcpy(&in, (char *)sin + 12, 4);
+		else	return;
+#else
+		in= *sin;
+#endif
+
+		he=gethostbyaddr( (char *)&in, sizeof(in), AF_INET);
 		if (!he)
 		{
 			switch (h_errno)	{
@@ -916,11 +1053,10 @@ char *r;
 ** check_blocklist is called once for each blocklist query to process.
 */
 
-static void check_blocklist(struct blocklist_s *p)
+static void docheckblocklist(struct blocklist_s *p, const char *nameptr)
 {
 const char *q;
 const char *varname=p->var;
-unsigned a,b,c,d;
 char	hostname[RFC1035_MAXNAMESIZE+1];
 char	buf[RFC1035_MAXNAMESIZE+1];
 int	hasnotxt;
@@ -928,22 +1064,14 @@ struct blocklist_s *pp;
 struct rfc1035_reply *replyp;
 int i;
 
+	hostname[0]=0;
+	strncat(hostname, nameptr, RFC1035_MAXNAMESIZE);
+
 	if (!varname)	varname="BLOCK";
 
 	if ((q=getenv(varname)) != 0 && *q)	return;
 					/* Env var already set */
 
-	/* Calculate DNS query hostname */
-
-	q=getenv("TCPREMOTEIP");
-
-	if (sscanf(q, "%u.%u.%u.%u", &a, &b, &c, &d) != 4)
-	{
-		return;
-	}
-
-	sprintf(hostname, "%u.%u.%u.%u.%s", d, c, b, a, p->zone);
-
 	if (p->ia.s_addr == INADDR_ANY)
 	{
 		/* We're not looking for a particular token IP address */
@@ -961,7 +1089,7 @@ int i;
 					RFC1035_CLASS_IN, &replyp)) >= 0)
 			{
 				rfc1035_rr_gettxt(replyp->allrrs[i], 0,
-						hostname);
+					hostname);
 				mysetenv(varname, hostname);
 			}
 		}
@@ -1070,11 +1198,53 @@ int i;
 	rfc1035_replyfree(replyp);
 }
 
+static void check_blocklist_ipv4(struct blocklist_s *p,
+	const struct in_addr *ia)
+{
+unsigned a,b,c,d;
+char	hostname[RFC1035_MAXNAMESIZE+1];
+const unsigned char *q=(const unsigned char *)ia;
+
+	/* Calculate DNS query hostname */
+
+	a=q[0];
+	b=q[1];
+	c=q[2];
+	d=q[3];
+
+	sprintf(hostname, "%u.%u.%u.%u.%s", d, c, b, a, p->zone);
+	docheckblocklist(p, hostname);
+}
+
+#if	RFC1035_IPV6
+
+static void check_blocklist(struct blocklist_s *p, const RFC1035_ADDR *ia)
+{
+	if (IN6_IS_ADDR_V4MAPPED(ia))
+	{
+	struct in_addr ia4;
+
+		memcpy(&ia4, (const char *)ia + 12, 4);
+		check_blocklist_ipv4(p, &ia4);
+	}
+}
+
+#else
+static void check_blocklist(struct blocklist_s *p, const RFC1035_ADDR *ia)
+{
+	check_blocklist_ipv4(p, ia);
+}
+#endif
+
 static void proxy();
 
-static void run(int fd, struct sockaddr_in *sin, const char *prog, char **argv)
+static void run(int fd, const RFC1035_ADDR *addr, int addrport,
+	const char *prog, char **argv)
 {
-struct sockaddr_in lsin;
+RFC1035_NETADDR lsin;
+RFC1035_ADDR laddr;
+int	lport;
+
 int	i;
 int	ipcnt, ccnt;
 char	buf[RFC1035_MAXNAMESIZE+128];
@@ -1083,15 +1253,16 @@ const char *remoteinfo;
 
 	i=sizeof(lsin);
 	if (sox_getsockname(fd, (struct sockaddr *)&lsin, &i) ||
-		i < sizeof(int) || lsin.sin_family != AF_INET)
+		rfc1035_sockaddrip(&lsin, i, &laddr) ||
+		rfc1035_sockaddrport(&lsin, i, &lport))
 	{
 		fprintf(stderr, "getsockname failed.\n");
 		exit(1);
 	}
 
 	if (!noidentlookup && (remoteinfo=tcpremoteinfo(
-		(struct sockaddr *)&lsin,
-		(struct sockaddr *)sin, 0)) != 0)
+		&laddr, lport,
+		addr, addrport, 0)) != 0)
 	{
 	char	*q=malloc(sizeof("TCPREMOTEINFO=")+strlen(remoteinfo));
 
@@ -1107,39 +1278,41 @@ const char *remoteinfo;
 
 	for (i=0, ipcnt=ccnt=0; i<nprocs; i++)
 	{
-	struct	sockaddr_in *psin;
+	RFC1035_ADDR *psin;
+	int	j;
 
 		if (pids[i] == (pid_t)-1)	continue;
 
-		psin=(struct sockaddr_in *)addrs+i;
-		if (sin->sin_family != AF_INET || psin->sin_family != AF_INET)
-			continue;
+		psin=addrs+i;
 
-		if ( sin->sin_addr.s_addr == psin->sin_addr.s_addr &&
+		for (j=0; j<sizeof(*addr); j++)
+			if ( ((char *)addr)[j] != ((char *)psin)[j])
+				break;
+
+		if (j >= sizeof(*addr) &&
 			++ipcnt >= maxperip)
 			_exit(0);	/* Too many from same IP address */
 
-		if ( (ntohl(sin->sin_addr.s_addr) & ~0xFF) ==
-			(ntohl(psin->sin_addr.s_addr) & ~0xFF) &&
+		if ( j >= sizeof(*addr)-1 &&
 			++ccnt >= maxperc)
 			_exit(0);	/* Too many from same netblock */
 
 	}
 
-	rfc1035_ntoa(sin->sin_addr, buf);
+	rfc1035_ntoa(addr, buf);
 	mysetenv("TCPREMOTEIP", buf);
-	sprintf(buf, "%d", ntohs(sin->sin_port));
+	sprintf(buf, "%d", ntohs(addrport));
 	mysetenv("TCPREMOTEPORT", buf);
-	ip2host(sin->sin_addr, "TCPREMOTEHOST");
+	ip2host(addr, "TCPREMOTEHOST");
 
-	rfc1035_ntoa(lsin.sin_addr, buf);
+	rfc1035_ntoa(&laddr, buf);
 	mysetenv("TCPLOCALIP", buf);
-	sprintf(buf, "%d", ntohs(lsin.sin_port));
+	sprintf(buf, "%d", ntohs(lport));
 	mysetenv("TCPLOCALPORT", buf);
-	ip2host(lsin.sin_addr, "TCPLOCALHOST");
+	ip2host(&laddr, "TCPLOCALHOST");
 
 	for (bl=blocklist; bl; bl=bl->next)
-		check_blocklist(bl);
+		check_blocklist(bl, addr);
 
 	sox_close(0);
 	sox_close(1);
